==> file4life-1.1.3-source.tgz: file4life-1.1.3/src/F4L/Actions/Storages.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Storages (
  fakeGetStorage
  , findStorage
  , forceGetEnvId
  , forceGetStorage
  , forceGetStorageId
  , storages
  , storagesForEnv
  , withStorage
  ) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Ops
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op

import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..) )
import Control.Monad.Reader
import Data.List( inits )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import System.FilePath( combine, dropTrailingPathSeparator, joinPath, splitDirectories )


forceGetEnvId :: ( FSMonad m ) => FilePath -> m EnvironmentId
forceGetEnvId dir = do
  f <- runErrorT $ readFromFile $ combine dir "env-id"
  case f of
    Right env -> return env
    Left _err -> do
      env <- generateUuid
      onErrorDie $ createDirectoryIfMissing False dir >> appendToFile ( combine dir "env-id" ) env
      return env


getStorageId :: ( FSMonad m, BufferMonad m ) => FilePath -> m ( Maybe StorageId )
getStorageId dir = do
  f <- runErrorT $ readFromFile $ combine dir storageIdFile
  return $ case f of
    Right sid -> Just sid
    Left _err -> Nothing


initStorageId :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                 => FilePath -> ErrorT Error ( F4LMonad k m ) StorageId
initStorageId path = do
  label <- lift $ do
    say Chatty $ "Init-ing storage on " ++ path
    env <- use st_envid
    let deflabel = path ++ " on " ++ env
    s <- interactivePrompt deflabel $ "Add label to created storage. Default is " ++ deflabel
    return $ if s == "" then deflabel else s
  sid <- lift generateUuid
  appendToFile ( combine path storageIdFile ) sid
  lift $ addOp $ AddStorageLabel sid label
  return sid


forceGetStorageId :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                     => FilePath -> ErrorT Error ( F4LMonad k m ) StorageId
forceGetStorageId path = do
  mpre <- lift $ getStorageId path
  case mpre of
    Just id -> return id
    Nothing -> initStorageId path


fakeGetStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                  => FilePath -> ErrorT Error ( F4LMonad k m ) RichPath
fakeGetStorage path = do
  canPath <- lift $ canonicalizePath path
  return $ RichPath "" canPath "" canPath


forceGetStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                   => FilePath -> ErrorT Error ( F4LMonad k m ) RichPath
forceGetStorage path = do
  say Chatty $ "Forcing storage on " ++ path
  mres <- lift $ getStorage path
  case mres of
    Just r -> say Timid ( "Found " ++ show r ) >> return r
    Nothing -> do
      canDir <- lift $ canonicalizePath path
      let paths = map joinPath $ drop 1 $ inits $ splitDirectories canDir
      let pathchoices = zip [1..] $ reverse paths
      dpath <- do
        s <- lift $ interactivePrompt "1" $ unlines $
             [ "No storage for " ++ path
             , "Create one" ] ++
             map ( \ (i, v) -> show i ++ " - " ++ v ) pathchoices
        return $ snd $ head $ filter ( ( s == ) . show . fst ) pathchoices
      _sid <- initStorageId dpath
      lift $ st_storages %= M.delete dpath
      forceGetStorage path


-- | returns storage, sPath, relPath
--   performs path canonicalization for sPath
getStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) => FilePath -> F4LMonad k m ( Maybe RichPath )
getStorage path = do
  smap <- use st_storages
  canPath <- canonicalizePath path
  let parts = splitDirectories canPath
  let dirs = reverse $ zipWith splitAt [ 1 .. length parts ] $ repeat parts
  msidpath <- testDirs smap dirs
  return $ liftM ( \ ( sid, sPath, relPath ) -> RichPath sid sPath relPath canPath ) msidpath
  where
    test smap dir = do
      case M.lookup dir smap of
        Just x @ ( Just _sid ) -> return x
        Just Nothing -> return Nothing
        Nothing -> do
          msid <- getStorageId dir
          st_storages %= M.insert dir msid
          return msid
    testDirs _ [] = return Nothing
    testDirs smap ( ( dparts, fparts ) : xs ) = do
      let dir = joinPath dparts
      res <- test smap dir
      case res of
        Nothing -> testDirs smap xs
        Just sid -> return $ Just ( sid, dir, joinPath fparts )


withStorage :: ( BufferMonad m, Knowledge k ) => StorageId -> FilePath -> F4LMonad k m ()
withStorage sid path = do
  oldS <- use $ st_opEnv . openv_storage
  when ( oldS /= Just ( sid, path ) ) $ addOp $ WithStorage sid $ dropTrailingPathSeparator path


storages :: ( FSMonad m, BufferMonad m, Knowledge k, SeenStoragesClass k ) 
            => GenF4LMonad k s m [(String, [String])]
storages = do
  m <- use $ st_knowledge . seen_storages
  return $ M.toList m


findStorage :: FSMonad m => StorageId -> [FilePath] -> m [FilePath]
findStorage sid paths = filterM isStorage paths
  where
    isStorage path = do
      esid <- runErrorT $ readFromFile $ combine path storageIdFile
      case esid of
        Right p | p == sid -> return True
        _ -> return False


storagesForEnv :: ( FSMonad m, KnownDataClass k )
                  => F4LMonad k m ( SM.Map StorageId [ FilePath ] )
storagesForEnv = do
  envid <- use st_envid
  say Timid $ "Env: " ++ envid
  -- Storages seen in this environment
  allStorages <- use $ st_knowledge . kd_storages
  let localStorages = storagesSeenIn envid allStorages
  liftM M.fromList $ mapM f $ M.toList localStorages
  where
    f (sid, paths) = liftM ( \ x -> (sid, x) ) $ findStorage sid paths

==> file4life-1.1.4-source.tgz: file4life-1.1.4/src/F4L/Actions/Storages.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Storages (
  fakeGetStorage
  , findStorage
  , forceGetEnvId
  , forceGetStorage
  , forceGetStorageId
  , storages
  , storagesForEnv
  , withStorage
  ) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Ops
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op

import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..) )
import Control.Monad.Reader
import Data.List( inits )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import System.FilePath( combine, dropFileName, dropTrailingPathSeparator, joinPath, splitDirectories )


forceGetEnvId :: ( FSMonad m ) => FilePath -> m EnvironmentId
forceGetEnvId dir = do
  f <- runErrorT $ readFromFile $ combine dir "env-id"
  case f of
    Right env -> return env
    Left _err -> do
      env <- generateUuid
      onErrorDie $ createDirectoryIfMissing False dir >> appendToFile ( combine dir "env-id" ) env
      return env


getStorageId :: ( FSMonad m, BufferMonad m ) => FilePath -> m ( Maybe StorageId )
getStorageId dir = do
  f <- runErrorT $ readFromFile $ combine dir storageIdFile
  return $ case f of
    Right sid -> Just sid
    Left _err -> Nothing


initStorageId :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                 => FilePath -> ErrorT Error ( F4LMonad k m ) StorageId
initStorageId path = do
  label <- lift $ do
    say Chatty $ "Init-ing storage on " ++ path
    env <- use st_envid
    let deflabel = path ++ " on " ++ env
    s <- interactivePrompt deflabel $ "Add label to created storage. Default is " ++ deflabel
    return $ if s == "" then deflabel else s
  sid <- lift generateUuid
  appendToFile ( combine path storageIdFile ) sid
  lift $ addOp $ AddStorageLabel sid label
  return sid


forceGetStorageId :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                     => FilePath -> ErrorT Error ( F4LMonad k m ) StorageId
forceGetStorageId path = do
  mpre <- lift $ getStorageId path
  case mpre of
    Just id -> return id
    Nothing -> initStorageId path


fakeGetStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                  => FilePath -> ErrorT Error ( F4LMonad k m ) RichPath
fakeGetStorage path = do
  canPath <- lift $ canonicalizePath path
  return $ RichPath "" canPath "" canPath


forceGetStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) 
                   => FilePath -> ErrorT Error ( F4LMonad k m ) RichPath
forceGetStorage path = do
  b <- lift $ doesFileExist path
  forceGetStorage' $ if b then dropFileName path else path
    where
      forceGetStorage' path = do
        say Chatty $ "Forcing storage on " ++ path
        mres <- lift $ getStorage path
        case mres of
          Just r -> say Timid ( "Found " ++ show r ) >> return r
          Nothing -> do
            canDir <- lift $ canonicalizePath path
            let paths = map joinPath $ drop 1 $ inits $ splitDirectories canDir
            let pathchoices = zip [1..] $ reverse paths
            dpath <- do
              s <- lift $ interactivePrompt "1" $ unlines $
                   [ "No storage for " ++ path
                   , "Create one" ] ++
                   map ( \ (i, v) -> show i ++ " - " ++ v ) pathchoices
              return $ snd $ head $ filter ( ( s == ) . show . fst ) pathchoices
            _sid <- initStorageId dpath
            lift $ st_storages %= M.delete dpath
            forceGetStorage path


-- | returns storage, sPath, relPath
--   performs path canonicalization for sPath
getStorage :: ( FSMonad m, BufferMonad m, Knowledge k ) => FilePath -> F4LMonad k m ( Maybe RichPath )
getStorage path = do
  smap <- use st_storages
  canPath <- canonicalizePath path
  let parts = splitDirectories canPath
  let dirs = reverse $ zipWith splitAt [ 1 .. length parts ] $ repeat parts
  msidpath <- testDirs smap dirs
  return $ liftM ( \ ( sid, sPath, relPath ) -> RichPath sid sPath relPath canPath ) msidpath
  where
    test smap dir = do
      case M.lookup dir smap of
        Just x @ ( Just _sid ) -> return x
        Just Nothing -> return Nothing
        Nothing -> do
          msid <- getStorageId dir
          st_storages %= M.insert dir msid
          return msid
    testDirs _ [] = return Nothing
    testDirs smap ( ( dparts, fparts ) : xs ) = do
      let dir = joinPath dparts
      res <- test smap dir
      case res of
        Nothing -> testDirs smap xs
        Just sid -> return $ Just ( sid, dir, joinPath fparts )


withStorage :: ( BufferMonad m, Knowledge k ) => StorageId -> FilePath -> F4LMonad k m ()
withStorage sid path = do
  oldS <- use $ st_opEnv . openv_storage
  when ( oldS /= Just ( sid, path ) ) $ addOp $ WithStorage sid $ dropTrailingPathSeparator path


storages :: ( FSMonad m, BufferMonad m, Knowledge k, SeenStoragesClass k ) 
            => GenF4LMonad k s m [(String, [String])]
storages = do
  m <- use $ st_knowledge . seen_storages
  return $ M.toList m


findStorage :: FSMonad m => StorageId -> [FilePath] -> m [FilePath]
findStorage sid paths = filterM isStorage paths
  where
    isStorage path = do
      esid <- runErrorT $ readFromFile $ combine path storageIdFile
      case esid of
        Right p | p == sid -> return True
        _ -> return False


storagesForEnv :: ( FSMonad m, KnownDataClass k )
                  => F4LMonad k m ( SM.Map StorageId [ FilePath ] )
storagesForEnv = do
  envid <- use st_envid
  say Timid $ "Env: " ++ envid
  -- Storages seen in this environment
  allStorages <- use $ st_knowledge . kd_storages
  let localStorages = storagesSeenIn envid allStorages
  liftM M.fromList $ mapM f $ M.toList localStorages
  where
    f (sid, paths) = liftM ( \ x -> (sid, x) ) $ findStorage sid paths
